#ifndef CORE_LIST_HPP
#define CORE_LIST_HPP

#include <assert.h>

#include "core/math/Math.hpp"
#include "core/utils/AlignedData.hpp"
#include "core/utils/ArrayString.hpp"
#include "core/utils/New.hpp"

namespace Core {
    template<typename T>
    class List final {
        size_t length;
        size_t capacity;
        T* data;

    public:
        List() : length(0), capacity(0), data(nullptr) {
        }

        List(const List& other)
            : length(0), capacity(other.capacity),
              data(allocate(other.capacity)) {
            for(const T& t : other) {
                unsafeAdd(t);
            }
        }

        List(List&& other) : List() {
            swap(other);
        }

        ~List() {
            clear();
            delete[] reinterpret_cast<AlignedType<T>*>(data);
        }

        List& operator=(List other) {
            swap(other);
            return *this;
        }

        T* begin() {
            return data;
        }

        T* end() {
            return data + length;
        }

        const T* begin() const {
            return data;
        }

        const T* end() const {
            return data + length;
        }

        void reserve(size_t n) {
            if(n > capacity) {
                setSize(n);
            }
        }

        void shrink() {
            if(length != capacity) {
                setSize(length);
            }
        }

        void resize(size_t n, const T& t) {
            if(length < n) {
                reserve(n);
                for(size_t i = n - length; i != 0; i--) {
                    unsafeAdd(t);
                }
            } else if(length > n) {
                for(size_t i = n; i < length; i++) {
                    data[i].~T();
                }
                length = n;
            }
        }

        void resize(size_t n) {
            if(length < n) {
                reserve(n);
                for(size_t i = n - length; i != 0; i--) {
                    unsafeAdd(T());
                }
            } else if(length > n) {
                for(size_t i = n; i < length; i++) {
                    data[i].~T();
                }
                length = n;
            }
        }

        template<typename... Args>
        T& put(Args&&... args) {
            ensureCapacity();
            return *unsafeAdd(Core::forward<Args>(args)...);
        }

        template<typename... Args>
        List& add(Args&&... args) {
            put(Core::forward<Args>(args)...);
            return *this;
        }

        T& operator[](size_t index) {
            return data[index];
        }

        const T& operator[](size_t index) const {
            return data[index];
        }

        size_t getLength() const {
            return length;
        }

        size_t getCapacity() const {
            return capacity;
        }

        void clear() {
            for(size_t i = 0; i < length; i++) {
                data[i].~T();
            }
            length = 0;
        }

        void removeBySwap(size_t index) {
            assert(index < length);
            length--;
            if(index != length) {
                data[index] = Core::move(data[length]);
            }
            data[length].~T();
        }

        void remove(size_t index) {
            assert(index < length);
            length--;
            T* currentT = begin() + index;
            T* endT = end();
            while(currentT != endT) {
                T* nextT = currentT + 1;
                *currentT = Core::move(*nextT);
                currentT = nextT;
            }
            endT->~T();
        }

        void removeLast() {
            removeBySwap(length - 1);
        }

        void toString(BufferString& s) const {
            Core::toString(s, *this);
        }

        void swap(List& other) {
            Core::swap(length, other.length);
            Core::swap(capacity, other.capacity);
            Core::swap(data, other.data);
        }

    private:
        static T* allocate(size_t n) {
            if(n <= 0) {
                return nullptr;
            }
            return reinterpret_cast<T*>(new(noThrow) AlignedType<T>[n]);
        }

        void ensureCapacity() {
            if(length >= capacity) {
                reserve(capacity + Math::max(4lu, capacity / 4));
            }
        }

        // does not check for capacity
        template<typename... Args>
        T* unsafeAdd(Args&&... args) {
            return new(data + length++) T(Core::forward<Args>(args)...);
        }

        void setSize(size_t n) {
            List copy;
            copy.data = allocate(n);
            copy.capacity = n;
            for(size_t i = 0; i < length; i++) {
                copy.unsafeAdd(Core::move(data[i]));
            }
            swap(copy);
        }
    };
}

#endif
